import glob

from cffi import FFI

ffibuilder = FFI()
ffibuilder.cdef(
    """
typedef uint8_t u8;
typedef uint16_t u16;
typedef uint32_t u32;
typedef uint64_t u64;

typedef int8_t s8;
typedef int16_t s16;
typedef int32_t s32;



#define BZ3_OK 0
#define BZ3_ERR_OUT_OF_BOUNDS -1
#define BZ3_ERR_BWT -2
#define BZ3_ERR_CRC -3
#define BZ3_ERR_MALFORMED_HEADER -4
#define BZ3_ERR_TRUNCATED_DATA -5
#define BZ3_ERR_DATA_TOO_BIG -6

struct bz3_state;

/**
 * @brief Get the last error number associated with a given state.
 */
int8_t bz3_last_error(struct bz3_state * state);

/**
 * @brief Return a user-readable message explaining the cause of the last error.
 */
const char * bz3_strerror(struct bz3_state * state);

/**
 * @brief Construct a new block encoder state, which will encode blocks as big as the given block size.
 * The decoder will be able to decode blocks at most as big as the given block size.
 * Returns NULL in case allocation fails or the block size is not between 65K and 511M
 */
struct bz3_state * bz3_new(int32_t block_size);

/**
 * @brief Free the memory occupied by a block encoder state.
 */
void bz3_free(struct bz3_state * state);

/**
 * @brief Return the recommended size of the output buffer for the compression functions.
 */
size_t bz3_bound(size_t input_size);

/* ** HIGH LEVEL APIs ** */

/**
 * @brief Compress a block of data. This function does not support parallelism
 * by itself, consider using the low level `bz3_encode_blocks()` function instead.
 * Using the low level API might provide better performance.
 * Returns a bzip3 error code; BZ3_OK when the operation is successful.
 * Make sure to set out_size to the size of the output buffer before the operation;
 * out_size must be at least equal to `bz3_bound(in_size)'.
 */
int bz3_compress(uint32_t block_size, const uint8_t * in, uint8_t * out, size_t in_size, size_t * out_size);

/**
 * @brief Decompress a block of data. This function does not support parallelism
 * by itself, consider using the low level `bz3_decode_blocks()` function instad.
 * Using the low level API might provide better performance.
 * Returns a bzip3 error code; BZ3_OK when the operation is successful.
 * Make sure to set out_size to the size of the output buffer before the operation.
 */
int bz3_decompress(const uint8_t * in, uint8_t * out, size_t in_size, size_t * out_size);

/**
 * @brief Encode a single block. Returns the amount of bytes written to `buffer'.
 * `buffer' must be able to hold at least `size + size / 50 + 32' bytes. The size must not
 * exceed the block size associated with the state.
 */
int32_t bz3_encode_block(struct bz3_state * state, uint8_t * buffer, int32_t size);

/**
 * @brief Decode a single block.
 * `buffer' must be able to hold at least `size + size / 50 + 32' bytes. The size must not exceed
 * the block size associated with the state.
 * @param size The size of the compressed data in `buffer'
 * @param orig_size The original size of the data before compression.
 */
int32_t bz3_decode_block(struct bz3_state * state, uint8_t * buffer, int32_t size, int32_t orig_size);

/**
 * @brief Encode `n' blocks, all in parallel.
 * All specifics of the `bz3_encode_block' still hold. The function will launch a thread for each block.
 * The compressed sizes are written to the `sizes' array. Every buffer is overwritten and none of them can overlap.
 * Precisely `n' states, buffers and sizes must be supplied.
 *
 * Expects `n' between 2 and 16.
 *
 * Present in the shared library only if -lpthread was present during building.
 */
//void bz3_encode_blocks(struct bz3_state * states[], uint8_t * buffers[], int32_t sizes[], int32_t n);

/**
 * @brief Decode `n' blocks, all in parallel.
 * Same specifics as `bz3_encode_blocks', but doesn't overwrite `sizes'.
 */
//void bz3_decode_blocks(struct bz3_state * states[], uint8_t * buffers[], int32_t sizes[], int32_t orig_sizes[], int32_t n);

const char * bz3_version();

s32 read_neutral_s32(u8 * data);
void write_neutral_s32(u8 * data, s32 value);
void* PyMem_Malloc(size_t n);
void PyMem_Free(void* p);
int strncmp (const char *s1, const char *s2, size_t size);
void *memcpy  (void *pto, const void *pfrom, size_t size);
    """
)

source = """
#include <stdint.h>
#include <string.h>
#include "common.h"
#include "libbz3.h"
#include "libsais.h"
"""
c_sources = glob.glob("./dep/src/*.c")
c_sources = list(filter(lambda x: "main" not in x, c_sources))
print(c_sources)

ffibuilder.set_source(
    "bz3.backends.cffi._bz3",
    source,
    sources=c_sources,
    include_dirs=["./dep/include"],
    define_macros=[("VERSION", '"1.3.1.r4-g6d041f4"')],
)

if __name__ == "__main__":
    ffibuilder.compile()



"""
Copyright (c) 2008-2023 synodriver <diguohuangjiajinweijun@gmail.com>
"""
import sys
import time
from concurrent.futures import ThreadPoolExecutor, wait
from unittest import TestCase

sys.path.append(".")

import bz3


class TestThread(TestCase):
    def test_class(self):
        start = time.time()
        futures = []
        with ThreadPoolExecutor(8) as pool:
            for i in range(1000):
                future = pool.submit(bz3.compress, b"1111" * 100000)
                futures.append(future)
            wait(futures)
            print("multithread done")
        print(f"{time.time() - start}")

    def test_single(self):
        start = time.time()
        for i in range(1000):
            bz3.compress(b"1111" * 100000)
        print("singlethread done")
        print(f"{time.time() - start}")


if __name__ == "__main__":
    import unittest

    unittest.main()



import sys
from typing import IO, Optional

from bz3.backends.cffi._bz3 import ffi, lib


def KiB(x: int) -> int:
    return x * 1024


def MiB(x: int) -> int:
    return x * 1024 * 1024


def check_file(file) -> bool:
    if hasattr(file, "read") and hasattr(file, "write"):
        return True
    return False


class BZ3Compressor:
    def __init__(self, block_size: int):
        if block_size < KiB(65) or block_size > MiB(511):
            raise ValueError("Block size must be between 65 KiB and 511 MiB")
        self.block_size = block_size
        self.state = lib.bz3_new(block_size)
        if self.state == ffi.NULL:
            raise MemoryError("Failed to create a block encoder state")
        self.buffer = ffi.cast(
            "uint8_t*", lib.PyMem_Malloc(lib.bz3_bound(block_size))
        )
        if self.buffer == ffi.NULL:
            lib.bz3_free(self.state)
            raise MemoryError("Failed to allocate memory")
        self.uncompressed = bytearray()
        self.have_magic_number = False  # 还没有写入magic number
        self.byteswap_buf = ffi.new(
            "uint8_t[4]"
        )  # fixme: multithread concurrent competition, e.g. single instance in multithread

    def __del__(self):
        if self.state != ffi.NULL:
            lib.bz3_free(self.state)
        if self.buffer != ffi.NULL:
            lib.PyMem_Free(self.buffer)

    def compress(self, data: bytes) -> bytes:
        input_size: int = len(data)
        ret = bytearray()
        if not self.have_magic_number:
            ret.extend(b"BZ3v1")
            lib.write_neutral_s32(
                ffi.cast("uint8_t*", self.byteswap_buf), self.block_size
            )
            ret.extend(ffi.unpack(ffi.cast("char*", self.byteswap_buf), 4))
            self.have_magic_number = True

        if input_size > 0:
            self.uncompressed.extend(data)
            while len(self.uncompressed) >= self.block_size:
                lib.memcpy(
                    self.buffer, ffi.from_buffer(self.uncompressed), self.block_size
                )
                # make a copy
                new_size = lib.bz3_encode_block(
                    self.state, self.buffer, self.block_size
                )
                if new_size == -1:
                    raise ValueError(
                        "Failed to encode a block: %s" % lib.bz3_strerror(self.state)
                    )

                lib.write_neutral_s32(ffi.cast("uint8_t*", self.byteswap_buf), new_size)
                ret.extend(ffi.unpack(ffi.cast("char*", self.byteswap_buf), 4))
                lib.write_neutral_s32(
                    ffi.cast("uint8_t*", self.byteswap_buf), self.block_size
                )
                ret.extend(ffi.unpack(ffi.cast("char*", self.byteswap_buf), 4))
                ret.extend(ffi.unpack(ffi.cast("char*", self.buffer), new_size))

                del self.uncompressed[: self.block_size]
        return bytes(ret)

    def flush(self) -> bytes:
        ret = bytearray()
        if self.uncompressed:
            lib.memcpy(
                self.buffer, ffi.from_buffer(self.uncompressed), len(self.uncompressed)
            )
            new_size = lib.bz3_encode_block(
                self.state, self.buffer, len(self.uncompressed)
            )
            if new_size == -1:
                raise ValueError(
                    "Failed to encode a block: %s" % lib.bz3_strerror(self.state)
                )
            # ret = PyBytes_FromStringAndSize(NULL, new_size + 8)
            # if not ret:
            #     raise
            lib.write_neutral_s32(ffi.cast("uint8_t*", self.byteswap_buf), new_size)
            ret.extend(ffi.unpack(ffi.cast("char*", self.byteswap_buf), 4))
            lib.write_neutral_s32(
                ffi.cast("uint8_t*", self.byteswap_buf), len(self.uncompressed)
            )
            ret.extend(ffi.unpack(ffi.cast("char*", self.byteswap_buf), 4))
            ret.extend(ffi.unpack(ffi.cast("char*", self.buffer), new_size))
            self.uncompressed.clear()
        return bytes(ret)

    def error(self) -> str:
        if lib.bz3_last_error(self.state) != lib.BZ3_OK:
            return ffi.string(lib.bz3_strerror(self.state)).decode()
        return None


class BZ3Decompressor:
    # cdef:
    #     bz3_state * state
    #     uint8_t * buffer
    #     int32_t block_size
    #     uint8_t byteswap_buf[4]
    #     bytearray unused  # 还没解压的数据
    #     bint have_magic_number

    # cdef readonly bint eof
    # cdef readonly bint needs_input
    # cdef readonly bint needs_input

    def init_state(self, block_size: int) -> int:
        """should exec only once"""
        self.block_size = block_size
        self.state = lib.bz3_new(block_size)
        if self.state == ffi.NULL:
            raise MemoryError("Failed to create a block encoder state")
        self.buffer = ffi.cast(
            "uint8_t*", lib.PyMem_Malloc(lib.bz3_bound(block_size))
        )
        if self.buffer == ffi.NULL:
            lib.bz3_free(self.state)
            self.state = ffi.NULL
            raise MemoryError("Failed to allocate memory")

    def __init__(self, ignore_error: bool = False):
        self.unused = bytearray()
        self.have_magic_number = False  # 还没有读到magic number
        self.ignore_error = ignore_error

    def __del__(self):
        if self.state != ffi.NULL:
            lib.bz3_free(self.state)
        if self.buffer != ffi.NULL:
            lib.PyMem_Free(self.buffer)

    def decompress(self, data: bytes) -> bytes:
        input_size: int = len(data)
        ret = bytearray()
        # cdef int32_t new_size, old_size, block_size
        if input_size > 0:
            # if PyByteArray_Resize(self.unused, input_size+PyByteArray_GET_SIZE(self.unused)) < 0:
            #     raise
            # memcpy(&(PyByteArray_AS_STRING(self.unused)[PyByteArray_GET_SIZE(self.unused)-input_size]), &data[0], input_size) # self.unused.extend
            self.unused.extend(data)
            if (
                len(self.unused) > 9 and not self.have_magic_number
            ):  # 9 bytes magic number
                if bytes(self.unused[:5]) != b"BZ3v1":
                    raise ValueError("Invalid signature")
                temp = self.unused[5:9]
                block_size = lib.read_neutral_s32(
                    ffi.cast("uint8_t*", ffi.from_buffer(temp))
                )
                if block_size < KiB(65) or block_size > MiB(511):
                    raise ValueError(
                        "The input file is corrupted. Reason: Invalid block size in the header"
                    )
                self.init_state(block_size)
                del self.unused[:9]
                self.have_magic_number = True

            while True:
                if len(self.unused) < 8:  # 8 byte的 header都不够 直接返回
                    break
                new_size = lib.read_neutral_s32(
                    ffi.cast("uint8_t*", ffi.from_buffer(self.unused))
                )  # todo gcc warning but bytes is contst
                temp = self.unused[4:8]
                old_size = lib.read_neutral_s32(
                    ffi.cast("uint8_t*", ffi.from_buffer(temp))
                )
                if old_size > lib.bz3_bound(self.block_size) or new_size > lib.bz3_bound(self.block_size):
                    raise ValueError("Failed to decode a block: Inconsistent headers.")
                if len(self.unused) < new_size + 8:  # 数据段不够
                    break
                temp = self.unused[8:]
                lib.memcpy(self.buffer, ffi.from_buffer(temp), new_size)

                code = lib.bz3_decode_block(self.state, self.buffer, new_size, old_size)
                if code == -1:
                    if self.ignore_error:
                        print(
                            f"Writing invalid block: {lib.bz3_strerror(self.state)}",
                            file=sys.stderr,
                        )
                    else:
                        raise ValueError(
                            "Failed to decode a block: %s"
                            % lib.bz3_strerror(self.state)
                        )
                ret.extend(ffi.unpack(ffi.cast("char*", self.buffer), old_size))
                del self.unused[: new_size + 8]
        return bytes(ret)

    @property
    def unused_data(self):
        """Data found after the end of the compressed stream."""
        return bytes(self.unused)

    def error(self) -> Optional[str]:
        if lib.bz3_last_error(self.state) != lib.BZ3_OK:
            return ffi.string(lib.bz3_strerror(self.state)).decode()
        return None


def compress_file(input: IO, output: IO, block_size: int) -> None:
    if not check_file(input):
        raise TypeError(
            "input except a file-like object, got %s" % type(input).__name__
        )
    if not check_file(output):
        raise TypeError(
            "output except a file-like object, got %s" % type(output).__name__
        )
    state = lib.bz3_new(block_size)
    if state == ffi.NULL:
        raise MemoryError("Failed to create a block encoder state")
    buffer = ffi.cast("uint8_t*", lib.PyMem_Malloc(lib.bz3_bound(block_size)))
    if buffer == ffi.NULL:
        lib.bz3_free(state)
        raise MemoryError
    byteswap_buf = ffi.new("uint8_t[4]")
    output.write(b"BZ3v1")
    lib.write_neutral_s32(byteswap_buf, block_size)
    output.write(ffi.unpack(ffi.cast("char*", byteswap_buf), 4))  # magic header

    try:
        while True:
            data = input.read(block_size)
            if not data:
                break
            lib.memcpy(buffer, ffi.from_buffer(data), len(data))
            new_size = lib.bz3_encode_block(state, buffer, len(data))
            if new_size == -1:
                raise ValueError(
                    "Failed to encode a block: %s" % lib.bz3_strerror(state)
                )
            lib.write_neutral_s32(byteswap_buf, new_size)
            output.write(ffi.unpack(ffi.cast("char*", byteswap_buf), 4))
            lib.write_neutral_s32(byteswap_buf, len(data))
            output.write(ffi.unpack(ffi.cast("char*", byteswap_buf), 4))
            output.write(ffi.unpack(ffi.cast("char*", buffer), new_size))
            output.flush()
    finally:
        output.flush()
        lib.bz3_free(state)
        lib.PyMem_Free(buffer)


def decompress_file(input: IO, output: IO) -> None:
    if not check_file(input):
        raise TypeError(
            "input except a file-like object, got %s" % type(input).__name__
        )
    if not check_file(output):
        raise TypeError(
            "output except a file-like object, got %s" % type(output).__name__
        )
    # cdef bytes data
    # cdef int32_t block_size
    data: bytes = input.read(9)  # magic and block_size type: bytes len = 9
    if len(data) < 9:
        raise ValueError("Invalid file. Reason: Smaller than magic header")
    if data[:5] != b"BZ3v1":
        raise ValueError("Invalid signature")
    block_size: int = lib.read_neutral_s32(
        ffi.cast("uint8_t*", ffi.from_buffer(data[5:]))
    )
    if block_size < KiB(65) or block_size > MiB(511):
        raise ValueError(
            "The input file is corrupted. Reason: Invalid block size in the header"
        )
    state = lib.bz3_new(block_size)
    if state == ffi.NULL:
        raise MemoryError("Failed to create a block encoder state")
    buffer = ffi.cast("uint8_t*", lib.PyMem_Malloc(lib.bz3_bound(block_size)))
    if buffer == ffi.NULL:
        lib.bz3_free(state)
        raise MemoryError("Failed to allocate memory")
    # cdef uint8_t byteswap_buf[4]
    # cdef int32_t new_size, old_size, code
    try:
        while True:
            data = input.read(4)
            if len(data) < 4:
                break
            new_size = lib.read_neutral_s32(ffi.cast("uint8_t*", ffi.from_buffer(data)))
            data = input.read(4)
            if len(data) < 4:
                break
            old_size = lib.read_neutral_s32(ffi.cast("uint8_t*", ffi.from_buffer(data)))
            if old_size > lib.bz3_bound(block_size) or new_size > lib.bz3_bound(block_size):
                raise ValueError("Failed to decode a block: Inconsistent headers.")
            data = input.read(new_size)  # type: bytes
            if len(data) < new_size:
                break
            lib.memcpy(buffer, ffi.cast("uint8_t*", ffi.from_buffer(data)), new_size)
            code = lib.bz3_decode_block(state, buffer, new_size, old_size)
            if code == -1:
                raise ValueError(
                    "Failed to decode a block: %s" % lib.bz3_strerror(state)
                )
            output.write(ffi.unpack(ffi.cast("char*", buffer), old_size))
            output.flush()
    finally:
        output.flush()
        lib.bz3_free(state)
        lib.PyMem_Free(buffer)


def recover_file(input: IO, output: IO) -> None:
    if not check_file(input):
        raise TypeError(
            "input except a file-like object, got %s" % type(input).__name__
        )
    if not check_file(output):
        raise TypeError(
            "output except a file-like object, got %s" % type(output).__name__
        )
    # cdef bytes data
    # cdef int32_t block_size
    data: bytes = input.read(9)  # magic and block_size type: bytes len = 9
    if len(data) < 9:
        raise ValueError("Invalid file. Reason: Smaller than magic header")
    if data[:5] != b"BZ3v1":
        raise ValueError("Invalid signature")
    block_size: int = lib.read_neutral_s32(
        ffi.cast("uint8_t*", ffi.from_buffer(data[5:]))
    )
    if block_size < KiB(65) or block_size > MiB(511):
        raise ValueError(
            "The input file is corrupted. Reason: Invalid block size in the header"
        )
    state = lib.bz3_new(block_size)
    if state == ffi.NULL:
        raise MemoryError("Failed to create a block encoder state")
    buffer = ffi.cast("uint8_t*", lib.PyMem_Malloc(lib.bz3_bound(block_size)))
    if buffer == ffi.NULL:
        lib.bz3_free(state)
        raise MemoryError("Failed to allocate memory")
    # cdef uint8_t byteswap_buf[4]
    # cdef int32_t new_size, old_size, code
    try:
        while True:
            data = input.read(4)
            if len(data) < 4:
                break
            new_size = lib.read_neutral_s32(ffi.cast("uint8_t*", ffi.from_buffer(data)))
            data = input.read(4)
            if len(data) < 4:
                break
            old_size = lib.read_neutral_s32(ffi.cast("uint8_t*", ffi.from_buffer(data)))
            if old_size > lib.bz3_bound(block_size) or new_size > lib.bz3_bound(block_size):
                raise ValueError("Failed to decode a block: Inconsistent headers.")
            data = input.read(new_size)  # type: bytes
            if len(data) < new_size:
                break
            lib.memcpy(buffer, ffi.cast("uint8_t*", ffi.from_buffer(data)), new_size)
            code = lib.bz3_decode_block(state, buffer, new_size, old_size)
            if code == -1:
                print(
                    f"Writing invalid block: {lib.bz3_strerror(state)}", file=sys.stderr
                )
            output.write(ffi.unpack(ffi.cast("char*", buffer), old_size))
            output.flush()
    finally:
        output.flush()
        lib.bz3_free(state)
        lib.PyMem_Free(buffer)


def test_file(input: IO, should_raise: bool = False) -> bool:
    if not check_file(input):
        raise TypeError(
            "input except a file-like object, got %s" % type(input).__name__
        )
    # cdef bytes data
    # cdef int32_t block_size
    data: bytes = input.read(9)  # magic and block_size type: bytes len = 9
    if len(data) < 9:
        if should_raise:
            raise ValueError("Invalid file. Reason: Smaller than magic header")
        return False
    if data[:5] != b"BZ3v1":
        if should_raise:
            raise ValueError("Invalid signature")
        return False
    block_size: int = lib.read_neutral_s32(
        ffi.cast("uint8_t*", ffi.from_buffer(data[5:]))
    )
    if block_size < KiB(65) or block_size > MiB(511):
        if should_raise:
            raise ValueError(
                "The input file is corrupted. Reason: Invalid block size in the header"
            )
        return False
    state = lib.bz3_new(block_size)
    if state == ffi.NULL:
        raise MemoryError("Failed to create a block encoder state")
    buffer = ffi.cast("uint8_t*", lib.PyMem_Malloc(lib.bz3_bound(block_size)))
    if buffer == ffi.NULL:
        lib.bz3_free(state)
        raise MemoryError("Failed to allocate memory")
    # cdef uint8_t byteswap_buf[4]
    # cdef int32_t new_size, old_size, code
    try:
        while True:
            data = input.read(4)
            if len(data) < 4:
                break
            new_size = lib.read_neutral_s32(ffi.cast("uint8_t*", ffi.from_buffer(data)))
            data = input.read(4)
            if len(data) < 4:
                break
            old_size = lib.read_neutral_s32(ffi.cast("uint8_t*", ffi.from_buffer(data)))
            if old_size > lib.bz3_bound(block_size) or new_size > lib.bz3_bound(block_size):
                raise ValueError("Failed to decode a block: Inconsistent headers.")
            data = input.read(new_size)  # type: bytes
            if len(data) < new_size:
                break
            lib.memcpy(buffer, ffi.cast("uint8_t*", ffi.from_buffer(data)), new_size)
            code = lib.bz3_decode_block(state, buffer, new_size, old_size)
            if code == -1:
                if should_raise:
                    raise ValueError(
                        "Failed to decode a block: %s" % lib.bz3_strerror(state)
                    )
                return False
        return True
    finally:
        lib.bz3_free(state)
        lib.PyMem_Free(buffer)


def bound(input_size: int) -> int:
    return lib.bz3_bound(input_size)


def compress_into(data, out, block_size: int = 1000000) -> int:
    out_size = ffi.new("size_t*")
    out_size[0] = len(out)
    bzerr = lib.bz3_compress(
        block_size, ffi.from_buffer(data), ffi.from_buffer(out), len(data), out_size
    )
    if bzerr != lib.BZ3_OK:
        raise ValueError(f"bz3_compress() failed with error code {bzerr}")
    return out_size[0]


def decompress_into(data, out) -> int:
    out_size = ffi.new("size_t*")
    out_size[0] = len(out)
    bzerr = lib.bz3_decompress(
        ffi.from_buffer(data), ffi.from_buffer(out), len(data), out_size
    )
    if bzerr != lib.BZ3_OK:
        raise ValueError(f"bz3_decompress() failed with error code {bzerr}")
    return out_size[0]


def libversion() -> str:
    return ffi.string(lib.bz3_version()).decode()



"""
Copyright (c) 2008-2023 synodriver <diguohuangjiajinweijun@gmail.com>
"""
import sys
import os
from unittest import TestCase

sys.path.append(".")

import memtrace
current_dir = os.path.dirname(__file__)


class TestSeek(TestCase):
    def test_not_leak(self):
        state = memtrace.State([(r"BZ3OmpDecompressor __cinit__ (?P<addr>\w+)",
                                 r"BZ3OmpDecompressor __dealloc__ (?P<addr>\w+)"),
                                (r"PyMem_Malloc (?P<addr>\w+)",
                                 r"PyMem_Free (?P<addr>\w+)"),
                                (r"bz3_new (?P<addr>\w+)",
                                 r"bz3_free (?P<addr>\w+)")
                                ])
        memtrace.parse_log(os.path.join(current_dir, "mem.log"), state)

    def test_not_leak2(self):
        state = memtrace.State([(r"BZ3OmpDecompressor __cinit__ (?P<addr>\w+)",
                                 r"BZ3OmpDecompressor __dealloc__ (?P<addr>\w+)"),
                                (r"PyMem_Malloc (?P<addr>\w+)",
                                 r"PyMem_Free (?P<addr>\w+)"),
                                (r"bz3_new (?P<addr>\w+)",
                                 r"bz3_free (?P<addr>\w+)")
                                ])
        memtrace.parse_log(os.path.join(current_dir, "mem2.log"), state)

    def test_not_leak3(self):
        state = memtrace.State([(r"BZ3OmpDecompressor __cinit__ (?P<addr>\w+)",
                                 r"BZ3OmpDecompressor __dealloc__ (?P<addr>\w+)"),
                                (r"PyMem_Malloc (?P<addr>\w+)",
                                 r"PyMem_Free (?P<addr>\w+)"),
                                (r"bz3_new (?P<addr>\w+)",
                                 r"bz3_free (?P<addr>\w+)")
                                ])
        memtrace.parse_log(os.path.join(current_dir, "mem3.log"), state)

    def tearDown(self) -> None:
        print("done")

if __name__ == "__main__":
    import unittest

    unittest.main()


"""
Copyright (c) 2008-2023 synodriver <diguohuangjiajinweijun@gmail.com>
"""
from bz3.backends.cython._bz3 import (
    BZ3Compressor,
    BZ3Decompressor,
    BZ3OmpCompressor,
    BZ3OmpDecompressor,
    bound,
    compress_file,
    compress_into,
    decompress_file,
    decompress_into,
    libversion,
    recover_file,
    test_file,
)



# cython: language_level=3
# cython: cdivision=True
cimport cython
from cpython.bytearray cimport PyByteArray_AS_STRING, PyByteArray_GET_SIZE
from cpython.bytes cimport (PyBytes_AS_STRING, PyBytes_FromStringAndSize,
                            PyBytes_GET_SIZE)
from cpython.mem cimport PyMem_Free, PyMem_Malloc
from cpython.object cimport PyObject_HasAttrString
from libc.stdint cimport int32_t, uint8_t, uint32_t
from libc.stdio cimport fprintf, stderr
from libc.string cimport memcpy, strncmp

from bz3.backends.cython.bzip3 cimport (BZ3_OK, KiB, MiB, bz3_bound,
                                        bz3_compress, bz3_decode_block,
                                        bz3_decompress, bz3_encode_block,
                                        bz3_free, bz3_last_error, bz3_new,
                                        bz3_state, bz3_strerror, bz3_version,
                                        read_neutral_s32, write_neutral_s32, MEMLOG)


cdef extern from "Python.h":
    void* PyMem_Calloc(size_t nelem, size_t elsize)

cdef const char* magic = "BZ3v1"

cdef inline uint8_t PyFile_Check(object file):
    if PyObject_HasAttrString(file, "read") and PyObject_HasAttrString(file, "write"):  # should we check seek method?
        return 1
    return 0

@cython.freelist(8)
@cython.no_gc
@cython.final
cdef class BZ3Compressor:
    cdef:
        bz3_state * state
        uint8_t * buffer
        readonly int32_t block_size
        bytearray uncompressed
        bint have_magic_number

    def __cinit__(self, int32_t block_size):
        if block_size < KiB(65) or block_size > MiB(511):
            raise ValueError("Block size must be between 65 KiB and 511 MiB")
        self.block_size = block_size
        self.state = bz3_new(block_size)
        if self.state == NULL:
            raise MemoryError("Failed to create a block encoder state")
        self.buffer = <uint8_t *>PyMem_Malloc(bz3_bound(block_size))
        if self.buffer == NULL:
            bz3_free(self.state)
            self.state = NULL
            raise MemoryError("Failed to allocate memory")
        self.uncompressed = bytearray()
        self.have_magic_number = 0 # 还没有写入magic number

    def __dealloc__(self):
        if self.state != NULL:
            bz3_free(self.state)
            self.state = NULL
        if self.buffer !=NULL:
            PyMem_Free(self.buffer)
            self.buffer = NULL

    cpdef inline bytes compress(self, const uint8_t[::1] data):
        cdef Py_ssize_t input_size = data.shape[0]
        cdef int32_t new_size
        cdef bytearray ret = bytearray()
        if not self.have_magic_number:
            # if PyByteArray_Resize(ret, 9) < 0:
            #     raise
            # memcpy(PyByteArray_AS_STRING(ret), magic, 5)
            ret.extend(<bytes>magic[:5]+b"\x00\x00\x00\x00")  # 9 bytes
            write_neutral_s32(<uint8_t*>&(PyByteArray_AS_STRING(ret)[5]), self.block_size)
            self.have_magic_number = 1

        if input_size > 0:
            # if PyByteArray_Resize(self.uncompressed, input_size+PyByteArray_GET_SIZE(self.uncompressed)) < 0:
            #     raise
            # memcpy(&(PyByteArray_AS_STRING(self.uncompressed)[PyByteArray_GET_SIZE(self.uncompressed)-input_size]), &data[0], input_size) # todo? direct copy to bytearray
            self.uncompressed.extend(data)
            while PyByteArray_GET_SIZE(self.uncompressed)>=self.block_size:
                memcpy(self.buffer, PyByteArray_AS_STRING(self.uncompressed), <size_t>self.block_size)
                # make a copy
                with nogil:
                    new_size = bz3_encode_block(self.state, self.buffer, self.block_size)
                if new_size == -1:
                    raise ValueError("Failed to encode a block: %s" % bz3_strerror(self.state))
                # if PyByteArray_Resize(ret, PyByteArray_GET_SIZE(ret) + new_size + 8) < 0:
                #     raise
                ret.extend((new_size + 8)*b"\x00")
                write_neutral_s32(<uint8_t*>&(PyByteArray_AS_STRING(ret)[PyByteArray_GET_SIZE(ret)-new_size-8]), new_size)
                write_neutral_s32(<uint8_t*>&(PyByteArray_AS_STRING(ret)[PyByteArray_GET_SIZE(ret)-new_size-4]), self.block_size)
                memcpy(&(PyByteArray_AS_STRING(ret)[PyByteArray_GET_SIZE(ret)-new_size]), self.buffer, <size_t>new_size)

                del self.uncompressed[:self.block_size]
        return bytes(ret)

    cpdef inline bytes flush(self):
        cdef bytes ret = b""
        cdef int32_t new_size
        cdef int32_t old_size = <int32_t>PyByteArray_GET_SIZE(self.uncompressed)
        if self.uncompressed:
            memcpy(self.buffer, PyByteArray_AS_STRING(self.uncompressed), <size_t>old_size)
            with nogil:
                new_size = bz3_encode_block(self.state, self.buffer, old_size)
            if new_size == -1:
                raise ValueError("Failed to encode a block: %s" % bz3_strerror(self.state))
            ret = PyBytes_FromStringAndSize(NULL, new_size + 8)
            if not ret:
                raise
            write_neutral_s32(<uint8_t*>PyBytes_AS_STRING(ret), new_size)
            write_neutral_s32(<uint8_t*>&(PyBytes_AS_STRING(ret)[4]), old_size)
            memcpy(&(PyBytes_AS_STRING(ret)[8]), self.buffer, <size_t> new_size)
            self.uncompressed.clear()
        return ret

    cpdef inline str error(self):
        if bz3_last_error(self.state) != BZ3_OK:
            return (<bytes>bz3_strerror(self.state)).decode()
        return None

@cython.freelist(8)
@cython.no_gc
@cython.final
cdef class BZ3Decompressor:
    cdef:
        bz3_state * state
        uint8_t * buffer
        readonly int32_t block_size
        bytearray unused  # 还没解压的数据
        bint have_magic_number
        readonly bint ignore_error # 是否忽略decode错误

    cdef inline int init_state(self, int32_t block_size) except -1:
        """should exec only once"""
        self.block_size = block_size
        self.state = bz3_new(block_size)
        if self.state == NULL:
            raise MemoryError("Failed to create a block encoder state")
        self.buffer = <uint8_t *> PyMem_Malloc(bz3_bound(block_size))
        if self.buffer == NULL:
            bz3_free(self.state)
            self.state = NULL
            raise MemoryError("Failed to allocate memory")

    def __cinit__(self, bint ignore_error = False):
        self.unused = bytearray()
        self.have_magic_number = 0 # 还没有读到magic number
        self.ignore_error = ignore_error

    def __dealloc__(self):
        if self.state != NULL:
            bz3_free(self.state)
            self.state = NULL
        if self.buffer !=NULL:
            PyMem_Free(self.buffer)
            self.buffer = NULL

    cpdef inline bytes decompress(self, const uint8_t[::1] data):
        cdef Py_ssize_t input_size = data.shape[0]
        cdef int32_t code
        cdef bytearray ret = bytearray()
        cdef int32_t new_size, old_size, block_size
        if input_size > 0:
            # if PyByteArray_Resize(self.unused, input_size+PyByteArray_GET_SIZE(self.unused)) < 0:
            #     raise
            # memcpy(&(PyByteArray_AS_STRING(self.unused)[PyByteArray_GET_SIZE(self.unused)-input_size]), &data[0], input_size) # self.unused.extend
            self.unused.extend(data)
            if PyByteArray_GET_SIZE(self.unused) > 9 and not self.have_magic_number: # 9 bytes magic number
                if strncmp(PyByteArray_AS_STRING(self.unused), magic, 5) != 0:
                    raise ValueError("Invalid signature")
                block_size = read_neutral_s32(<uint8_t*>&(PyByteArray_AS_STRING(self.unused)[5]))
                if block_size  < KiB(65) or block_size >MiB(511):
                    raise ValueError("The input file is corrupted. Reason: Invalid block size in the header")
                self.init_state(block_size)
                del self.unused[:9]
                self.have_magic_number = 1

            while True:
                if PyByteArray_GET_SIZE(self.unused)<8: # 8 byte的 header都不够 直接返回
                    break
                new_size = read_neutral_s32(<uint8_t*>PyByteArray_AS_STRING(self.unused)) # todo gcc warning but bytes is contst
                old_size = read_neutral_s32(<uint8_t*>&(PyByteArray_AS_STRING(self.unused)[4]))
                if old_size > <int32_t>bz3_bound(self.block_size) or new_size > <int32_t>bz3_bound(self.block_size):
                    raise ValueError("Failed to decode a block: Inconsistent headers.")
                if PyByteArray_GET_SIZE(self.unused) < new_size+8: # 数据段不够
                    break
                memcpy(self.buffer, &(PyByteArray_AS_STRING(self.unused)[8]), <size_t>new_size)
                with nogil:
                    code = bz3_decode_block(self.state, self.buffer, new_size, old_size)
                if code == -1:
                    if self.ignore_error:
                        fprintf(stderr, "Writing invalid block: %s\n", bz3_strerror(self.state))
                    else:
                        raise ValueError("Failed to decode a block: %s" % bz3_strerror(self.state))
                # if PyByteArray_Resize(ret, PyByteArray_GET_SIZE(ret) + old_size) < 0:
                #     raise
                ret.extend(<bytes>self.buffer[:old_size])
                # memcpy(&(PyByteArray_AS_STRING(ret)[PyByteArray_GET_SIZE(ret)-old_size]), self.buffer, <size_t>old_size)
                del self.unused[:new_size+8]
        return bytes(ret)

    @property
    def unused_data(self):
        """Data found after the end of the compressed stream."""
        return bytes(self.unused)

    cpdef inline str error(self):
        if bz3_last_error(self.state) != BZ3_OK:
            return (<bytes> bz3_strerror(self.state)).decode()
        return None


def compress_file(object input, object output, int32_t block_size):
    if not PyFile_Check(input):
        raise TypeError("input except a file-like object, got %s" % type(input).__name__)
    if not PyFile_Check(output):
        raise TypeError("output except a file-like object, got %s" % type(output).__name__)
    cdef bz3_state *state = bz3_new(block_size)
    if state == NULL:
        raise MemoryError("Failed to create a block encoder state")
    cdef uint8_t * buffer = <uint8_t *>PyMem_Malloc(bz3_bound(block_size))
    if buffer == NULL:
        bz3_free(state)
        state = NULL
        raise MemoryError
    cdef bytes data
    cdef int32_t new_size
    cdef uint8_t byteswap_buf[4]

    output.write(b"BZ3v1")
    write_neutral_s32(byteswap_buf, block_size)
    output.write(PyBytes_FromStringAndSize(<char*>&byteswap_buf[0], 4))  # magic header
    cdef int32_t old_size
    try:
        while True:
            data = input.read(block_size)
            if not data:
                break
            old_size = <int32_t>PyBytes_GET_SIZE(data)
            memcpy(buffer, PyBytes_AS_STRING(data), <size_t>old_size)
            with nogil:
                new_size = bz3_encode_block(state, buffer, old_size)
            if new_size == -1:
                raise ValueError("Failed to encode a block: %s" % bz3_strerror(state))
            write_neutral_s32(byteswap_buf, new_size)
            output.write(PyBytes_FromStringAndSize(<char*>&byteswap_buf[0], 4))
            write_neutral_s32(byteswap_buf, old_size)
            output.write(PyBytes_FromStringAndSize(<char*>&byteswap_buf[0], 4))
            output.write(PyBytes_FromStringAndSize(<char*>buffer, new_size))
            output.flush()
    finally:
        output.flush()
        bz3_free(state)
        state = NULL
        PyMem_Free(buffer)
        buffer = NULL

def decompress_file(object input, object output):
    if not PyFile_Check(input):
        raise TypeError("input except a file-like object, got %s" % type(input).__name__)
    if not PyFile_Check(output):
        raise TypeError("output except a file-like object, got %s" % type(output).__name__)
    cdef bytes data
    cdef int32_t block_size
    data = input.read(9) # magic and block_size type: bytes len = 9
    if PyBytes_GET_SIZE(data) < 9:
        raise ValueError("Invalid file. Reason: Smaller than magic header")
    if strncmp(PyBytes_AS_STRING(data), magic, 5) != 0:
        raise ValueError("Invalid signature")
    block_size = read_neutral_s32(<uint8_t*>&(PyBytes_AS_STRING(data)[5]))
    if block_size < KiB(65) or block_size > MiB(511):
        raise ValueError("The input file is corrupted. Reason: Invalid block size in the header")
    cdef bz3_state *state = bz3_new(block_size)
    if state == NULL:
        raise MemoryError("Failed to create a block encoder state")
    cdef uint8_t *buffer = <uint8_t *> PyMem_Malloc(bz3_bound(block_size))
    if buffer == NULL:
        bz3_free(state)
        state = NULL
        raise MemoryError("Failed to allocate memory")
    cdef int32_t new_size, old_size, code

    try:
        while True:
            data = input.read(4)
            if PyBytes_GET_SIZE(data) < 4:
                break
            new_size = read_neutral_s32(<uint8_t*>PyBytes_AS_STRING(data))
            data = input.read(4)
            if PyBytes_GET_SIZE(data) < 4:
                break
            old_size = read_neutral_s32(<uint8_t *> PyBytes_AS_STRING(data))
            if old_size > <int32_t>bz3_bound(block_size) or new_size > <int32_t>bz3_bound(block_size):
                raise ValueError("Failed to decode a block: Inconsistent headers.")
            data = input.read(new_size) # type: bytes
            if PyBytes_GET_SIZE(data) < new_size:
                break
            memcpy(buffer, PyBytes_AS_STRING(data), <size_t> new_size)
            with nogil:
                code = bz3_decode_block(state, buffer, new_size, old_size)
            if code == -1:
                raise ValueError("Failed to decode a block: %s" % bz3_strerror(state))
            output.write(PyBytes_FromStringAndSize(<char*>buffer, old_size))
            output.flush()
    finally:
        output.flush()
        bz3_free(state)
        state = NULL
        PyMem_Free(buffer)
        buffer = NULL

def recover_file(object input, object output):
    if not PyFile_Check(input):
        raise TypeError("input except a file-like object, got %s" % type(input).__name__)
    if not PyFile_Check(output):
        raise TypeError("output except a file-like object, got %s" % type(output).__name__)
    cdef bytes data
    cdef int32_t block_size
    data = input.read(9) # magic and block_size type: bytes len = 9
    if PyBytes_GET_SIZE(data) < 9:
        raise ValueError("Invalid file. Reason: Smaller than magic header")
    if strncmp(PyBytes_AS_STRING(data), magic, 5) != 0:
        raise ValueError("Invalid signature")
    block_size = read_neutral_s32(<uint8_t*>&(PyBytes_AS_STRING(data)[5]))
    if block_size < KiB(65) or block_size > MiB(511):
        raise ValueError("The input file is corrupted. Reason: Invalid block size in the header")
    cdef bz3_state *state = bz3_new(block_size)
    if state == NULL:
        raise MemoryError("Failed to create a block encoder state")
    cdef uint8_t *buffer = <uint8_t *> PyMem_Malloc(bz3_bound(block_size))
    if buffer == NULL:
        bz3_free(state)
        state = NULL
        raise MemoryError("Failed to allocate memory")
    cdef int32_t new_size, old_size, code

    try:
        while True:
            data = input.read(4)
            if PyBytes_GET_SIZE(data) < 4:
                break
            new_size = read_neutral_s32(<uint8_t*>PyBytes_AS_STRING(data))
            data = input.read(4)
            if PyBytes_GET_SIZE(data) < 4:
                break
            old_size = read_neutral_s32(<uint8_t *> PyBytes_AS_STRING(data))
            if old_size > <int32_t>bz3_bound(block_size) or new_size > <int32_t>bz3_bound(block_size):
                raise ValueError("Failed to decode a block: Inconsistent headers.")
            data = input.read(new_size) # type: bytes
            if PyBytes_GET_SIZE(data) < new_size:
                break
            memcpy(buffer, PyBytes_AS_STRING(data), <size_t> new_size)
            with nogil:
                code = bz3_decode_block(state, buffer, new_size, old_size)
            if code == -1:
                fprintf(stderr, "Writing invalid block: %s\n", bz3_strerror(state))
            output.write(PyBytes_FromStringAndSize(<char*>buffer, old_size))
            output.flush()
    finally:
        output.flush()
        bz3_free(state)
        state = NULL
        PyMem_Free(buffer)
        buffer = NULL

cpdef inline bint test_file(object input, bint should_raise = False) except? 0:
    if not PyFile_Check(input):
        raise TypeError("input except a file-like object, got %s" % type(input).__name__)
    cdef bytes data
    cdef int32_t block_size
    data = input.read(9)  # magic and block_size type: bytes len = 9
    if PyBytes_GET_SIZE(data) < 9:
        if should_raise:
            raise ValueError("Invalid file. Reason: Smaller than magic header")
        return 0
    if strncmp(PyBytes_AS_STRING(data), magic, 5) != 0:
        if should_raise:
            raise ValueError("Invalid signature")
        return 0
    block_size = read_neutral_s32(<uint8_t *> &(PyBytes_AS_STRING(data)[5]))
    if block_size < KiB(65) or block_size > MiB(511):
        if should_raise:
            raise ValueError("The input file is corrupted. Reason: Invalid block size in the header")
        return 0
    cdef bz3_state *state = bz3_new(block_size)
    if state == NULL:
        raise MemoryError("Failed to create a block encoder state")
    cdef uint8_t *buffer = <uint8_t *> PyMem_Malloc(bz3_bound(block_size))
    if buffer == NULL:
        bz3_free(state)
        state = NULL
        raise MemoryError("Failed to allocate memory")
    cdef int32_t new_size, old_size, code

    try:
        while True:
            data = input.read(4)
            if PyBytes_GET_SIZE(data) < 4:
                break
            new_size = read_neutral_s32(<uint8_t *> PyBytes_AS_STRING(data))
            data = input.read(4)
            if PyBytes_GET_SIZE(data) < 4:
                break
            old_size = read_neutral_s32(<uint8_t *> PyBytes_AS_STRING(data))
            if old_size > <int32_t>bz3_bound(block_size) or new_size > <int32_t>bz3_bound(block_size):
                if should_raise:
                    raise ValueError("Failed to decode a block: Inconsistent headers.")
                return 0
            data = input.read(new_size)  # type: bytes
            if PyBytes_GET_SIZE(data) < new_size:
                break
            memcpy(buffer, PyBytes_AS_STRING(data), <size_t> new_size)
            with nogil:
                code = bz3_decode_block(state, buffer, new_size, old_size)
            # print(f"newsize {new_size} oldsize {old_size}") # todo
            if code == -1:
                if should_raise:
                    raise ValueError("Failed to decode a block: %s" % bz3_strerror(state))
                return 0
        return 1
    finally:
        bz3_free(state)
        state = NULL
        PyMem_Free(buffer)
        buffer = NULL

cpdef inline size_t bound(size_t input_size) nogil:
    return bz3_bound(input_size)


cpdef inline size_t compress_into(const uint8_t[::1] data, uint8_t[::1] out, uint32_t block_size = 1000000) except? 0:
    cdef:
        size_t out_size = <size_t>out.shape[0]
        int bzerr
    with nogil:
        bzerr = bz3_compress(block_size, &data[0], &out[0], <size_t>data.shape[0], &out_size)
    if bzerr != BZ3_OK:
        raise ValueError(f"bz3_compress() failed with error code {bzerr}")
    return out_size

cpdef inline size_t decompress_into(const uint8_t[::1] data, uint8_t[::1] out) except? 0:
    cdef:
        size_t out_size = <size_t>out.shape[0]
        int bzerr
    with nogil:
        bzerr = bz3_decompress(&data[0], &out[0], <size_t>data.shape[0], &out_size)
    if bzerr != BZ3_OK:
        raise ValueError(f"bz3_decompress() failed with error code {bzerr}")
    return out_size

cpdef inline str libversion():
    return (<bytes>bz3_version()).decode()

# openmp
from cython.parallel cimport prange


cdef void bz3_encode_blocks(bz3_state ** states, uint8_t ** buffers, int32_t *sizes, int32_t numthreads):
    # sizes: read and write
    cdef int32_t i
    for i in prange(numthreads, nogil=True, schedule="static", num_threads=numthreads):
        sizes[i] = bz3_encode_block(states[i], buffers[i], sizes[i])


@cython.freelist(8)
@cython.no_gc
@cython.final
cdef class BZ3OmpCompressor:
    cdef:
        bz3_state ** states
        uint8_t ** buffers
        int32_t * sizes   # compressed
        int32_t * old_sizes # origin size
        readonly int32_t block_size
        bytearray uncompressed
        bint have_magic_number
        readonly uint32_t numthreads  # how many threads to use

    def __cinit__(self, int32_t block_size, uint32_t numthreads):
        if block_size < KiB(65) or block_size > MiB(511):
            raise ValueError("Block size must be between 65 KiB and 511 MiB")
        self.block_size = block_size
        self.states = <bz3_state **>PyMem_Calloc(<size_t>numthreads, sizeof(bz3_state *)) # prepare the array
        if not self.states:
            raise MemoryError
        MEMLOG("PyMem_Malloc %p\n", self.states)
        self.buffers = <uint8_t **>PyMem_Calloc(<size_t>numthreads, sizeof(uint8_t *))
        if not self.buffers:
            PyMem_Free(self.states)
            self.states = NULL
            raise MemoryError
        MEMLOG("PyMem_Malloc %p\n", self.buffers)
        self.sizes = <int32_t *>PyMem_Malloc(sizeof(int32_t) * numthreads)
        if not self.sizes:
            PyMem_Free(self.states)
            MEMLOG("PyMem_Free %p\n", self.states)
            self.states = NULL
            PyMem_Free(self.buffers)
            MEMLOG("PyMem_Free %p\n", self.buffers)
            self.buffers = NULL
            raise MemoryError
        MEMLOG("PyMem_Malloc %p\n", self.sizes)
        self.old_sizes = <int32_t *> PyMem_Malloc(sizeof(int32_t) * numthreads)
        if not self.old_sizes:
            PyMem_Free(self.states)
            MEMLOG("PyMem_Free %p\n", self.states)
            self.states = NULL
            PyMem_Free(self.buffers)
            MEMLOG("PyMem_Free %p\n", self.buffers)
            self.buffers = NULL
            PyMem_Free(self.sizes)
            MEMLOG("PyMem_Free %p\n", self.sizes)
            self.sizes = NULL
            raise MemoryError
        MEMLOG("PyMem_Malloc %p\n", self.old_sizes)

        cdef uint32_t i
        try:
            for i in range(numthreads):
                self.states[i] = bz3_new(block_size)
                if self.states[i] == NULL:
                    raise MemoryError("Failed to create a block encoder state")  # todo 如何善后
                MEMLOG("bz3_new %p\n", self.states[i])
                self.buffers[i] = <uint8_t *>PyMem_Malloc(bz3_bound(block_size))
                if self.buffers[i] == NULL:
                    raise MemoryError("Failed to allocate memory")
                MEMLOG("PyMem_Malloc %p\n", self.buffers[i])
        except:
            self.free_states()
            self.free_buffers()
            PyMem_Free(self.states)
            MEMLOG("PyMem_Free %p\n", self.states)
            self.states = NULL
            PyMem_Free(self.buffers)
            MEMLOG("PyMem_Free %p\n", self.buffers)
            self.buffers = NULL
            PyMem_Free(self.sizes)
            MEMLOG("PyMem_Free %p\n", self.sizes)
            self.sizes = NULL
            PyMem_Free(self.old_sizes)
            MEMLOG("PyMem_Free %p\n", self.old_sizes)
            self.old_sizes = NULL
            raise
        self.uncompressed = bytearray()
        self.have_magic_number = 0 # 还没有写入magic number
        self.numthreads = numthreads

    cdef inline void free_states(self):
        cdef uint32_t i
        if self.states:
            for i in range(self.numthreads):
                if self.states[i]:
                    bz3_free(self.states[i])
                    MEMLOG("bz3_free %p\n", self.states[i])
                    self.states[i] = NULL

    cdef inline void free_buffers(self):
        cdef uint32_t i
        if self.buffers:
            for i in range(self.numthreads):
                if self.buffers[i]:
                    PyMem_Free(self.buffers[i])
                    MEMLOG("PyMem_Free %p\n", self.buffers[i])
                    self.buffers[i] = NULL

    def __dealloc__(self):
        cdef uint32_t i
        self.free_states()
        self.free_buffers()
        if self.states:
            PyMem_Free(self.states)
            MEMLOG("PyMem_Free %p\n", self.states)
            self.states = NULL
        if self.buffers:
            PyMem_Free(self.buffers)
            MEMLOG("PyMem_Free %p\n", self.buffers)
            self.buffers = NULL
        if self.sizes:
            PyMem_Free(self.sizes)
            MEMLOG("PyMem_Free %p\n", self.sizes)
            self.sizes = NULL
        if self.old_sizes:
            PyMem_Free(self.old_sizes)
            MEMLOG("PyMem_Free %p\n", self.old_sizes)
            self.old_sizes = NULL

    cpdef inline bytes compress(self, const uint8_t[::1] data):
        cdef Py_ssize_t input_size = data.shape[0]
        cdef int32_t new_size
        cdef bytearray ret = bytearray()
        cdef int32_t all_blocks_size = self.block_size * self.numthreads
        if not self.have_magic_number:
            # if PyByteArray_Resize(ret, 9) < 0:
            #     raise
            # memcpy(PyByteArray_AS_STRING(ret), magic, 5)
            ret.extend(<bytes>magic[:5]+b"\x00\x00\x00\x00")  # 9 bytes
            write_neutral_s32(<uint8_t*>&(PyByteArray_AS_STRING(ret)[5]), self.block_size)
            self.have_magic_number = 1
        cdef uint32_t i
        if input_size > 0:
            # if PyByteArray_Resize(self.uncompressed, input_size+PyByteArray_GET_SIZE(self.uncompressed)) < 0:
            #     raise
            # memcpy(&(PyByteArray_AS_STRING(self.uncompressed)[PyByteArray_GET_SIZE(self.uncompressed)-input_size]), &data[0], input_size) # todo? direct copy to bytearray
            self.uncompressed.extend(data)
            if PyByteArray_GET_SIZE(self.uncompressed) >= all_blocks_size:  # able to perform a compress
                while PyByteArray_GET_SIZE(self.uncompressed) >= all_blocks_size:  # ensure fill all blocks
                    for i in range(self.numthreads):
                        self.sizes[i] = self.block_size  # fill the sizes array
                        memcpy(self.buffers[i], &PyByteArray_AS_STRING(self.uncompressed)[i*self.block_size], <size_t>self.block_size)
                        # make a copy
                    bz3_encode_blocks(self.states, self.buffers, self.sizes, <int32_t>self.numthreads)
                    for i in range(self.numthreads):
                        if bz3_last_error(self.states[i]) != BZ3_OK:
                            raise ValueError("Failed to encode a block: %s" % bz3_strerror(self.states[i]))
                        # if PyByteArray_Resize(ret, PyByteArray_GET_SIZE(ret) + new_size + 8) < 0:
                        #     raise
                        ret.extend((self.sizes[i] + 8)*b"\x00")
                        write_neutral_s32(<uint8_t*>&(PyByteArray_AS_STRING(ret)[PyByteArray_GET_SIZE(ret)-self.sizes[i]-8]), self.sizes[i])
                        write_neutral_s32(<uint8_t*>&(PyByteArray_AS_STRING(ret)[PyByteArray_GET_SIZE(ret)-self.sizes[i]-4]), self.block_size)
                        memcpy(&(PyByteArray_AS_STRING(ret)[PyByteArray_GET_SIZE(ret)-self.sizes[i]]), self.buffers[i], <size_t>self.sizes[i])

                    del self.uncompressed[:all_blocks_size]
            return bytes(ret)

    cpdef inline bytes flush(self):
        cdef bytearray ret = bytearray()
        cdef int32_t new_size
        cdef int32_t remain_size = <int32_t>PyByteArray_GET_SIZE(self.uncompressed)
        cdef:
            int i = 0  # thread count
            int j
        if self.uncompressed:  # will perform a compress
            while self.block_size * (i+1) < remain_size:
                memcpy(self.buffers[i],
                       &PyByteArray_AS_STRING(self.uncompressed)[i*self.block_size],
                       <size_t>self.block_size)
                self.sizes[i] = self.old_sizes[i] = self.block_size
                # old_sizes[i] = self.block_size
                i += 1
            memcpy(self.buffers[i],
                   &PyByteArray_AS_STRING(self.uncompressed)[i * self.block_size],
                   <size_t> (remain_size-i*self.block_size))  # fill as many blocks as possible
            self.sizes[i] = self.old_sizes[i] = remain_size-i*self.block_size
            # old_sizes[i] = remain_size-i*self.block_size
            i += 1
            bz3_encode_blocks(self.states, self.buffers, self.sizes, i)
            for j in range(i):  # state index
                if bz3_last_error(self.states[j]) != BZ3_OK:
                    raise ValueError("Failed to encode a block: %s" % bz3_strerror(self.states[j]))

                ret.extend((self.sizes[j] + 8) * b"\x00")
                write_neutral_s32(<uint8_t *> &(PyByteArray_AS_STRING(ret)[PyByteArray_GET_SIZE(ret) - self.sizes[j] - 8]),
                                  self.sizes[j])
                write_neutral_s32(<uint8_t *> &(PyByteArray_AS_STRING(ret)[PyByteArray_GET_SIZE(ret) - self.sizes[j] - 4]),
                                  self.old_sizes[j])
                memcpy(&(PyByteArray_AS_STRING(ret)[PyByteArray_GET_SIZE(ret) - self.sizes[j]]), self.buffers[j],
                       <size_t> self.sizes[j])

            self.uncompressed.clear()
        return bytes(ret)

    cpdef inline list error(self):
        cdef uint32_t i
        cdef list ret = []
        for i in range(self.numthreads):
            if bz3_last_error(self.states[i]) != BZ3_OK:
                ret.append((<bytes>bz3_strerror(self.states[i])).decode())
        return ret


cdef void bz3_decode_blocks(bz3_state ** states, uint8_t ** buffers, int32_t* sizes, int32_t* orig_size, int32_t numthreads):
    cdef int32_t i
    for i in prange(numthreads, nogil=True, schedule='static', num_threads=numthreads):
        bz3_decode_block(states[i], buffers[i], sizes[i], orig_size[i])


@cython.freelist(8)
@cython.no_gc
@cython.final
cdef class BZ3OmpDecompressor:
    cdef:
        bz3_state ** states
        uint8_t ** buffers
        int32_t * sizes   # compressed
        int32_t * old_sizes  # origin
        readonly int32_t block_size
        bytearray unused  # 还没解压的数据
        bint have_magic_number
        readonly uint32_t numthreads  # how many threads to use
        readonly bint ignore_error  # 是否忽略decode错误

    cdef inline int init_state(self, int32_t block_size) except -1:
        """should exec only once"""
        if not self.states:
            self.states = <bz3_state **> PyMem_Calloc(self.numthreads, sizeof(bz3_state *))  # prepare the array
            if not self.states:
                raise MemoryError
            MEMLOG("PyMem_Malloc %p\n", self.states)
        if not self.buffers:
            self.buffers = <uint8_t **> PyMem_Calloc(self.numthreads, sizeof(uint8_t *))
            if not self.buffers:
                raise MemoryError
            MEMLOG("PyMem_Malloc %p\n", self.buffers)
        cdef uint32_t i
        try:
            for i in range(self.numthreads):
                self.states[i] = bz3_new(block_size)
                if self.states[i] == NULL:
                    raise MemoryError("Failed to create a block encoder state")  # todo 如何善后
                MEMLOG("bz3_new %p\n", self.states[i])
                self.buffers[i] = <uint8_t *> PyMem_Malloc(bz3_bound(block_size))
                if self.buffers[i] == NULL:
                    raise MemoryError("Failed to allocate memory")
                MEMLOG("PyMem_Malloc %p\n", self.buffers[i])
        except:
            self.free_states()
            self.free_buffers()
            raise
        self.block_size = block_size

    def __cinit__(self,  uint32_t numthreads, bint ignore_error = False):
        self.states = NULL
        self.buffers = NULL
        self.unused = bytearray()
        self.have_magic_number = 0 # 还没有读到magic number
        self.numthreads = numthreads
        self.ignore_error = ignore_error

        self.sizes = <int32_t *> PyMem_Malloc(sizeof(int32_t) * numthreads)
        if not self.sizes:
            raise MemoryError
        MEMLOG("PyMem_Malloc %p\n", self.sizes)
        self.old_sizes = <int32_t *> PyMem_Malloc(sizeof(int32_t) * numthreads)
        if not self.old_sizes:
            PyMem_Free(self.sizes)
            self.sizes = NULL
            raise MemoryError
        MEMLOG("PyMem_Malloc %p\n", self.old_sizes)
        MEMLOG("BZ3OmpDecompressor __cinit__ %p\n", <void*>self)

    cdef inline void free_states(self):
        cdef uint32_t i
        if self.states: # states数组是否初始化，即是否调用过init_states
            for i in range(self.numthreads):
                if self.states[i]:
                    bz3_free(self.states[i])
                    MEMLOG("bz3_free %p\n", self.states[i])
                    self.states[i] = NULL

    cdef inline void free_buffers(self):
        cdef uint32_t i # states数组是否初始化，即是否调用过init_states
        if self.buffers: # states数组是否初始化，即是否调用过init_states
            for i in range(self.numthreads):
                if self.buffers[i]:
                    PyMem_Free(self.buffers[i])
                    MEMLOG("PyMem_Free %p\n", self.buffers[i])
                    self.buffers[i] = NULL

    def __dealloc__(self):
        self.free_states()
        self.free_buffers()
        if self.states:
            PyMem_Free(self.states)
            MEMLOG("PyMem_Free %p\n", self.states)
            self.states = NULL
        if self.buffers:
            PyMem_Free(self.buffers)
            MEMLOG("PyMem_Free %p\n", self.buffers)
            self.buffers = NULL
        if self.sizes:
            PyMem_Free(self.sizes)
            MEMLOG("PyMem_Free %p\n", self.sizes)
            self.sizes = NULL
        if self.old_sizes:
            PyMem_Free(self.old_sizes)
            MEMLOG("PyMem_Free %p\n", self.old_sizes)
            self.old_sizes = NULL
        MEMLOG("BZ3OmpDecompressor __dealloc__ %p\n", <void *> self)

    cpdef inline bytes decompress(self, const uint8_t[::1] data):
        cdef Py_ssize_t input_size = data.shape[0]
        cdef int32_t code
        cdef bytearray ret = bytearray()
        cdef int32_t  block_size
        cdef uint32_t i, thread_count, j, should_delete=0
        cdef int should_break = 0
        if input_size > 0:
            # if PyByteArray_Resize(self.unused, input_size+PyByteArray_GET_SIZE(self.unused)) < 0:
            #     raise
            # memcpy(&(PyByteArray_AS_STRING(self.unused)[PyByteArray_GET_SIZE(self.unused)-input_size]), &data[0], input_size) # self.unused.extend
            self.unused.extend(data) # read header
            if PyByteArray_GET_SIZE(self.unused) > 9 and not self.have_magic_number: # 9 bytes magic number
                if strncmp(PyByteArray_AS_STRING(self.unused), magic, 5) != 0:
                    raise ValueError("Invalid signature")
                block_size = read_neutral_s32(<uint8_t*>&(PyByteArray_AS_STRING(self.unused)[5]))
                if block_size  < KiB(65) or block_size >MiB(511):
                    raise ValueError("The input file is corrupted. Reason: Invalid block size in the header")
                self.init_state(block_size)
                del self.unused[:9]
                self.have_magic_number = 1
            # 有几个block就用几个
            while not should_break:
                thread_count = 0  # 这一波能用上几个thread
                # should_delete = 0 # 讀取完成后從self.unused刪多少
                for i in range(self.numthreads):
                    if (PyByteArray_GET_SIZE(self.unused)-should_delete)<8: # 8 byte的 header都不够 直接返回
                        should_break = 1
                        break
                    self.sizes[i] = read_neutral_s32(<uint8_t*>&PyByteArray_AS_STRING(self.unused)[should_delete]) # todo gcc warning but bytes is const
                    self.old_sizes[i] = read_neutral_s32(<uint8_t*>&(PyByteArray_AS_STRING(self.unused)[should_delete+4]))
                    if self.old_sizes[i] > <int32_t>bz3_bound(self.block_size) or self.sizes[i] > <int32_t>bz3_bound(self.block_size):
                        raise ValueError("Failed to decode a block: Inconsistent headers.")
                    if (PyByteArray_GET_SIZE(self.unused)-should_delete) < self.sizes[i]+8: # 数据段不够
                        should_break = 1
                        break
                    memcpy(self.buffers[i], &(PyByteArray_AS_STRING(self.unused)[should_delete+8]), <size_t>self.sizes[i])
                    should_delete += (self.sizes[i] + 8)
                    thread_count += 1
                if thread_count:  # 一个block都凑不齐decode个jb
                    bz3_decode_blocks(self.states, self.buffers, self.sizes, self.old_sizes, <int32_t>thread_count)
                for j in range(thread_count):
                    if bz3_last_error(self.states[j]) != BZ3_OK:
                        if self.ignore_error:
                            fprintf(stderr, "Writing invalid block: %s\n", bz3_strerror(self.states[j]))
                        else:
                            raise ValueError("Failed to decode data: %s" % bz3_strerror(self.states[j]))
                    ret.extend(<bytes>self.buffers[j][:self.old_sizes[j]])
            if should_delete:
                del self.unused[:should_delete]
        return bytes(ret)

    @property
    def unused_data(self):
        """Data found after the end of the compressed stream."""
        return bytes(self.unused)

    cpdef inline list error(self):
        cdef uint32_t i
        cdef list ret = []
        for i in range(self.numthreads):
            if bz3_last_error(self.states[i]) != BZ3_OK:
                ret.append((<bytes> bz3_strerror(self.states[i])).decode())
        return ret



"""
Copyright (c) 2008-2023 synodriver <diguohuangjiajinweijun@gmail.com>
"""
__version__ = "0.1.3rc1"

from bz3.backends import (
    bound,
    compress_file,
    compress_into,
    decompress_file,
    decompress_into,
    libversion,
    recover_file,
    test_file,
)
from bz3.bz3 import BZ3File, compress, decompress, open



"""
Copyright (c) 2008-2023 synodriver <diguohuangjiajinweijun@gmail.com>
"""
import sys

sys.path.append(".")
import os
from random import randint
from unittest import TestCase

from bz3 import (
    bound,
    compress_file,
    compress_into,
    decompress_file,
    decompress_into,
    libversion,
)
from bz3 import open as bz3_open
from bz3 import test_file

# os.environ["BZ3_USE_CFFI"] = "1"


class TestCompress(TestCase):
    def test_compress(self):
        with open("test_input.tar", "rb") as inp, open("compressed.bz3", "wb") as out:
            compress_file(inp, out, 1000 * 1000)

    def test_decompress(self):
        with open("compressed.bz3", "rb") as inp, open("output.tar", "wb") as out:
            decompress_file(inp, out)

    def test_filelike_write(self):
        with bz3_open("test.bz3", "wt", encoding="utf-8") as f:
            f.write("test data")
        with bz3_open("test.bz3", "rt", encoding="utf-8") as f:
            self.assertEqual(f.read(), "test data")

    def test_filelike_seek(self):
        with bz3_open("test.bz3", "wt", encoding="utf-8") as f:
            f.write("test data")
        with bz3_open("test.bz3", "rb") as f:
            f.seek(0, 2)
            f.seek(0, 0)
            self.assertEqual(f.read(), b"test data")

    def test_zerocopy(self):
        outsize = bound(100)
        out = bytearray(200)
        out2 = bytearray(200)
        for i in range(1000):
            inp = bytes([randint(0, 255) for _ in range(100)])
            buffer_updated = compress_into(inp, out)
            buffer_updated = decompress_into(out[:buffer_updated], out2)
            self.assertEqual(bytes(out2[:buffer_updated]), inp)

    def test_version(self):
        self.assertTrue(isinstance(libversion(), str))


if __name__ == "__main__":
    import unittest

    unittest.main()



"""
Copyright (c) 2008-2023 synodriver <diguohuangjiajinweijun@gmail.com>
"""
import os
import platform

impl = platform.python_implementation()


def _should_use_cffi() -> bool:
    ev = os.getenv("BZ3_USE_CFFI")
    if ev is not None:
        return True
    if impl == "CPython":
        return False
    else:
        return True


if not _should_use_cffi():
    from bz3.backends.cython import (
        BZ3Compressor,
        BZ3Decompressor,
        BZ3OmpCompressor,
        BZ3OmpDecompressor,
        bound,
        compress_file,
        compress_into,
        decompress_file,
        decompress_into,
        libversion,
        recover_file,
        test_file,
    )
else:
    from bz3.backends.cffi import (
        BZ3Compressor,
        BZ3Decompressor,
        bound,
        compress_file,
        compress_into,
        decompress_file,
        decompress_into,
        libversion,
        recover_file,
        test_file,
    )



# cython: language_level=3
# cython: cdivision=True
from libc.stdint cimport (int8_t, int16_t, int32_t, uint8_t, uint16_t,
                          uint32_t, uint64_t)


cdef extern from "common.h" nogil:
    int32_t KiB(int32_t x)
    int32_t MiB(int32_t x)
    ctypedef uint8_t u8
    ctypedef uint16_t u16
    ctypedef uint32_t u32
    ctypedef uint64_t u64

    ctypedef int8_t s8
    ctypedef int16_t s16
    ctypedef int32_t s32

    s32 read_neutral_s32(const u8 * data)
    void write_neutral_s32(u8 * data, s32 value)

cdef extern from "libbz3.h" nogil:
    int8_t BZ3_OK
    int8_t BZ3_ERR_OUT_OF_BOUNDS
    int8_t BZ3_ERR_BWT
    int8_t BZ3_ERR_CRC
    int8_t BZ3_ERR_MALFORMED_HEADER
    int8_t BZ3_ERR_TRUNCATED_DATA
    int8_t BZ3_ERR_DATA_TOO_BIG

    struct bz3_state:
        pass

    int8_t bz3_last_error(bz3_state * state)
    const char * bz3_strerror(bz3_state * state)
    bz3_state * bz3_new(int32_t block_size)
    void bz3_free(bz3_state * state)
    size_t bz3_bound(size_t input_size)
    int bz3_compress(uint32_t block_size, const uint8_t * in_, uint8_t * out, size_t in_size, size_t * out_size)
    int bz3_decompress(const uint8_t * in_, uint8_t * out, size_t in_size, size_t * out_size)
    int32_t bz3_encode_block(bz3_state * state, uint8_t * buffer, int32_t size)
    int32_t bz3_decode_block(bz3_state * state, uint8_t * buffer, int32_t size, int32_t orig_size)
    void bz3_encode_blocks(bz3_state * states[], uint8_t * buffers[], int32_t sizes[], int32_t n)
    void bz3_decode_blocks(bz3_state * states[], uint8_t * buffers[], int32_t sizes[], int32_t orig_sizes[],
                           int32_t n)
    const char * bz3_version()

cdef extern from "<stdio.h>" nogil:
    """
#ifdef MEMDEBUG
    #define MEMLOG(...) fprintf(stderr, __VA_ARGS__)
#else
    #define MEMLOG(...)
#endif
    """
    void MEMLOG(const char* fmt, ...)


import sys
from unittest import TestCase

sys.path.append(".")

import os

from bz3 import compress, decompress
from bz3 import open as _open
from bz3.backends.cython import BZ3OmpCompressor, BZ3OmpDecompressor

origin_data = b"124" * (100 * 10**6 + 7)


class TestOmp(TestCase):
    def test_compress(self):
        compressor = BZ3OmpCompressor(10**6, os.cpu_count())
        # origin_data = b"124" * (600*10 ** 6 + 7)
        compressed_data = compressor.compress(origin_data) + compressor.flush()
        self.assertEqual(decompress(compressed_data), origin_data)

    def test_decompress(self):
        # origin_data = b"124" * (10 ** 6 + 7)
        compressed_data = compress(origin_data)
        decompressor = BZ3OmpDecompressor(os.cpu_count())
        self.assertEqual(decompressor.decompress(compressed_data), origin_data)

    def test_stream(self):
        with _open("test.bz3", "wb", num_threads=16) as f:
            f.write(origin_data)
        with _open("test.bz3", "rb", num_threads=16) as f:
            self.assertEqual(f.read(), origin_data)


if __name__ == "__main__":
    import unittest

    unittest.main()



"""
Copyright (c) 2008-2023 synodriver <diguohuangjiajinweijun@gmail.com>
"""
import sys
import os
from unittest import TestCase

sys.path.append(".")

import bz3

current_dir = os.path.dirname(__file__)


class TestSeek(TestCase):
    def setUp(self) -> None:
        if not os.path.exists(os.path.join(current_dir, "test.bz3")):
            with bz3.open(os.path.join(current_dir, "test.bz3"), "wb", num_threads=os.cpu_count()) as f:
                f.write(b"1111" * 1000)

    def test_seek(self):
        with bz3.open(os.path.join(current_dir, "test.bz3"), "rb", num_threads=os.cpu_count()) as f:
            for i in range(100):
                self.assertEqual(f.read(1000), b"1" * 1000, "read error")
                f.seek(0, 2)
                f.seek(0, 0)
            pass

    def tearDown(self) -> None:
        print("done")

if __name__ == "__main__":
    import unittest

    unittest.main()
